import os
import sys
import re
import shutil
import glob

import pandas as pd

from Bio import SeqIO
from collections import Counter

'''
Read GTDB-tk taxonomy and define metagenomic species
Collect all genes
'''

# Make directories
try:
    os.mkdir(os.path.join(snakemake.input[1], 'genes'))
except FileExistsError:
    pass
try:
    os.mkdir(os.path.join(snakemake.input[1], 'tabs'))
except FileExistsError:
    pass

def most_common(ll):
    return(Counter(ll).most_common(1)[0])

empty_tax = ['d__', 'p__', 'c__', 'o__', 'f__', 'g__', 's__']

# Get consensus annotation for clusters
tax_list = []
gene_path_list = []
gene_path_prot_list = []
for clust in os.listdir(snakemake.input[0]):

    # Try reading both bacterial and archaeal summaries.
    try:
        tax_bac = pd.read_csv(glob.glob(os.path.join(snakemake.input[0], clust, 'gtdbtk.bac*.summary.tsv'))[0], sep='\t', header=0)  
        if tax_bac.iloc[0,1]=='Unclassified':
            tax_bac=None
        elif tax_bac.iloc[0,1]=='Unclassified Bacteria':
            tax_bac=None
    except (IndexError, FileNotFoundError):
        tax_bac = None
    try: 
        tax_ar = pd.read_csv(glob.glob(os.path.join(snakemake.input[0], clust, 'gtdbtk.ar*.summary.tsv'))[0], sep='\t', header=0)  
    except (IndexError, FileNotFoundError):
        tax_ar = None

    # Combine
    try:
        tax = pd.concat([tax_bac, tax_ar])
    except Exception:
        tax = None

    if tax is not None:
        
        classification = [x.split(';') for x in tax['classification']]
        
        # Remove unclassified
        classification = [x for x in classification if len(x) == 7]
        
        # Traverse from species annotation and up
        # Pick the annotation if the most common annotation is above prevalence set by parameter
        level = 6
        while level >= 0:
            annot = [x[level] for x in classification]
            cons = most_common(annot)
            if cons[1] >= float(snakemake.params[0])*len(annot):
                rep_annot = classification[annot.index(cons[0])]
                # If annotation not species level, add empty annotations for consistency
                while level < 6:
                    rep_annot[level+1] = empty_tax[level+1]
                    level += 1
                final_annot = ';'.join(rep_annot)
                break
            else:
                level -= 1

        tax_list.append([[clust], final_annot])

    # Collect gene sequence paths
    gene_path = os.path.join(snakemake.input[0], clust, 'identify', 'intermediate_results', 'marker_genes')

    for bin_id in os.listdir(gene_path):
        gene_path_list.append(os.path.join(gene_path, bin_id, bin_id+'_protein.fna'))
        gene_path_prot_list.append(os.path.join(gene_path, bin_id, bin_id+'_protein.faa'))

# If similar at species level, then aggregate
species_list = []
mgs_list = []
for clust in tax_list:
    # If vamb_cluster parameter is not set, aggregate if species annotation is similar
    if bool(snakemake.params[2]):
        # If not species annotation, just add to final list
        if bool(re.search('s__$', clust[1])):
            mgs_list.append(clust)
        # If species annotation, check if it's already added to list - if False add to final list. If True append to existing
        else:
            if clust[1] in species_list:
                mgs_list[[x[1] for x in mgs_list].index(clust[1])][0] += clust[0]
            else:
                mgs_list.append(clust)
                species_list.append(clust[1])
    # vamb_cluster parameter is set, clusters are not aggregated
    else:
        mgs_list.append(clust)

# Write metagenomicspecies file
with open(snakemake.output[0], 'w') as fh:
    for mgs in mgs_list:
        rep = min([int(x) for x in mgs[0]])
        fh.write('{}\t{}\t{}\n'.format(str(rep), mgs[1], ','.join(mgs[0])))

# Write genes
with open(snakemake.output[1], 'wb') as wfh:
    for f in gene_path_list:
        with open(f, 'rb') as rfh:
            shutil.copyfileobj(rfh, wfh)

#write genes as proteins
with open(snakemake.output[5], 'wb') as wfh:
    for f in gene_path_prot_list:
        with open(f, 'rb') as rfh:
            shutil.copyfileobj(rfh, wfh)


# Collect marker gene info for phylogenetic analyses
## Get contig-bin info
with open(snakemake.params[1], 'r') as fh:
    vamb = [x.strip().split() for x in fh.readlines()]
vamb = {x[1]: x[0] for x in vamb}

## Find all GTDB marker files
with open(snakemake.output[2], 'w') as wfh:
    for f in glob.glob(os.path.join(snakemake.input[0], '*', 'identify', 'intermediate_results', 'marker_genes', '*', '*tophit.tsv')):
        with open(f, 'r') as rfh:
            nl = 0
            for ll in rfh:
                if nl > 0:
                    line = ll.strip().split()
                    wfh.write(line[0]+'\t'+re.sub(',.*', '', line[1])+'\t'+vamb[re.sub('_[0-9]*$', '', line[0])]+'\n')
                nl += 1

# Collect all unique markers for phylogenetic analyses
def unique_gtdb(domain, output):
    with open(output, 'w') as wfh:
        for f in glob.glob(os.path.join(snakemake.input[0], '*', 'identify', 'gtdbtk.'+domain+'*.markers_summary.tsv')):
            with open(f, 'r') as rfh:
                nl = 0
                for ll in rfh:
                    if nl > 0:
                        line = ll.strip().split()
                        wfh.write(line[0]+'\t'+line[5]+'\n')
                    nl += 1

unique_gtdb('bac', snakemake.output[3])
unique_gtdb('ar', snakemake.output[4])



#!/usr/bin/env python

import argparse
import logging
import os
import sys
import pkg_resources

from os.path import join, dirname, basename
from shutil import copyfile
from argparse import RawTextHelpFormatter

from maginator.controller import Controller
from maginator.workflow import Workflow

# Snakefiles
_ROOT = os.path.abspath(os.path.dirname(__file__))
WORKFLOW_FILTER = os.path.join(_ROOT, 'workflow', 'filter.Snakefile')
WORKFLOW_GTDBTK = os.path.join(_ROOT, 'workflow', 'gtdbtk.Snakefile')
WORKFLOW_PARSE_GTDBTK = os.path.join(_ROOT, 'workflow', 'parse_gtdbtk.Snakefile')
WORKFLOW_FILTER_GENE_CLUSTERS = os.path.join(_ROOT, 'workflow', 'filter_geneclusters.Snakefile')
WORKFLOW_GENE_COUNT_MAT = os.path.join(_ROOT, 'workflow', 'gene_count_mat.Snakefile')
WORKFLOW_PRESCREENING_GENES = os.path.join(_ROOT, 'workflow', 'prescreening_genes.Snakefile')
WORKFLOW_SIGNATURE_GENES = os.path.join(_ROOT, 'workflow', 'signature_genes.Snakefile')
WORKFLOW_OUTGROUP = os.path.join(_ROOT, 'workflow', 'outgroup.Snakefile')
WORKFLOW_PILEUP = os.path.join(_ROOT, 'workflow', 'pileup.Snakefile')
WORKFLOW_PILEUP_PARSE = os.path.join(_ROOT, 'workflow', 'pileup_parse.Snakefile')
WORKFLOW_ALIGNMENT = os.path.join(_ROOT, 'workflow', 'alignment.Snakefile')
WORKFLOW_PHYLO = os.path.join(_ROOT, 'workflow', 'phylo.Snakefile')
WORKFLOW_GENE_TAX = os.path.join(_ROOT, 'workflow', 'gene_tax.Snakefile')

def cli():
    
    ########## Arguments ##########
    ap = argparse.ArgumentParser(description='MAGinator version {}'.format(pkg_resources.require("maginator")[0].version), add_help=False, allow_abbrev=False, formatter_class=RawTextHelpFormatter)
    
    # Required
    apr = ap.add_argument_group('required arguments')
    apr.add_argument('-v', '--vamb_clusters', help='Path to VAMB clusters.tsv file', required=True)
    apr.add_argument('-r', '--reads', help='Comma-delimited file with format: SampleName,AbsolutePathToForwardReads,AbsolutePathToReverseReads. SampleNames should match the 1st column in clusters.tsv with the pattern SampleName_{clusternumber}', required=True)
    apr.add_argument('-c', '--contigs', help='Fasta file with contig sequences. Fasta headers should match the 2nd column in the clusters.tsv file', required=True)
    apr.add_argument('-o', '--output', help='Prefix for output directory', required=True)
    apr.add_argument('-g', '--gtdb_db', help='Path to GTDB-tk database', type=str, required=True)

    # Cluster arguments
    apc = ap.add_argument_group('compute cluster arguments')
    apc.add_argument('--cluster', help='Cluster compute structure [%(default)s]', default=None, type=str, choices=[None,'qsub','slurm','drmaa'])
    apc.add_argument('--cluster_info', help='Cluster scheduler arguments when submitting cluster jobs.\nHas to contain the following special strings:\n{mem_gb}, {cores}, and {runtime}.\nThese special strings will be substituted by maginator to indicate resources for each job.\n{mem_gb} is substituted for the mem_gb in GB.\n{runtime} is substituted with the time in the following format: DD:HH:MM:SS.\nCan also contain user names, groups, etc. required by the cluster scheduler', default=None, type=str)
    apc.add_argument('--max_jobs', help='Maximum number of cluster jobs [%(default)s]', default=500, type=int)
    
    # Optional
    apo = ap.add_argument_group('optional arguments')
    apo.add_argument('-V','--version', action='version', version='MAGinator version {}'.format(pkg_resources.require("maginator")[0].version))
    apo.add_argument("-h", "--help", action="help", help="show this help message and exit")
    apo.add_argument('--max_cores', help='Maximum number of cores [%(default)s]', default=40, type=int)
    apo.add_argument('--max_mem', help='Maximum mem_gb in GB [%(default)s]', default=180, type=int)
    apo.add_argument('--log_lvl', help='Logging level [%(default)s].', default='INFO', type=str, choices=['DEBUG','INFO','WARNING','ERROR'])
    apo.add_argument('--only_conda', help='Only install conda environments, then exit', action='store_true')
    apo.add_argument('--snake', help='Only run specific snakemake command. For debug purposes', type=str)
    apo.add_argument('--unlock', help='Unlock snakemake directory in case of unexpected exists, then exit', action='store_true')

    # Parameters
    app = ap.add_argument_group('parameters')
    app.add_argument('--binsize', help='Minimum bin size for inclusion [%(default)s].', default=200000, type=int)
    app.add_argument('--mgs_collections', help='If set, bin clusters will be aggregated to metagenomic species.', action='store_true')
    app.add_argument('--annotation_prevalence', help='Minimum prevalence of taxonomic assignment in a cluster of bins to call consensus [%(default)s]', default=0.75, type=float)
    app.add_argument('--clustering_coverage', help='Alignment coverage for clustering of genes with MMseqs2 [%(default)s]', default=0.8, type=float)
    app.add_argument('--clustering_min_seq_id', help='Sequence identity threshold for clustering of genes with MMseqs2 [%(default)s]', default=0.95, type=float)
    app.add_argument('--clustering_type', help='Sequence type for gene clustering with MMseqs2. Nucleotide- or protein-level [%(default)s]', default='protein', type=str, choices=['nucleotide', 'protein'])
    app.add_argument('--min_gtdb_markers', help='Minimum GTDBtk marker genes shared between MGS and outgroup for rooting trees [%(default)s]', default=10, type=int)
    app.add_argument('--marker_gene_cluster_prevalence', help='Minimum prevalence of marker genes to be selected for rooting of MGS trees [%(default)s]', default=0.5, type=float)

    app.add_argument('--min_mapped_signature_genes', help='Minimum number of signature genes with reads mapped for the sample to be included in the refinement [%(default)s]', default=3, type=int)
    app.add_argument('--min_samples', help='Minimum number of samples containing the MAG cluster (more than "min_mapped_signature_genes" present) for the MAG cluster to identidy SG [%(default)s]', default=3, type=int)

    app.add_argument('--min_af', help='Minimim allele frequency for calling a base when creating phylogenies [%(default)s]', default=0.8, type=float)
    app.add_argument('--min_depth', help='Minimim read depth for calling a base when creating phylogenies [%(default)s]', default=2, type=int)
    app.add_argument('--min_nonN', help='Minimum fraction of non-N characters of a sample to be included in a phylogeny [%(default)s]', default=0.5, type=float)
    app.add_argument('--min_marker_genes', help='Minimum marker genes to be detected for inclusion of a sample in a phylogeny [%(default)s]', default=2, type=int)
    app.add_argument('--min_signature_genes', help='Minimum signature genes to be detected for inclusion of a sample in a phylogeny [%(default)s]', default=50, type=int)
    app.add_argument('--af_cutoff', help='Cutoff for average median allele frequency of SG alignment for determining mixed/pure population of a MAG in a sample [%(default)s]', default=0, type=float)
    app.add_argument('--phylo', help='Software for phylogeny inference. Either fast (fasttree) or slow and more accurate (iqtree) [%(default)s]', default='fasttree', type=str, choices=['fasttree', 'iqtree'])
    app.add_argument('--tax_scope_threshold', help='Threshold for assigning the taxonomic scope of a gene cluster [%(default)s]', default=0.9, type=float)
    app.add_argument('--synteny_adj_cutoff', help='Minimum number of times gene clusters should be adjacent to be included in synteny graph [%(default)s]', default=1, type=int)
    app.add_argument('--synteny_mcl_inflation', help='Inflation parameter for mcl clustering of synteny graph. Usually between 1.2 and 5. Higher values produce smaller clusters [%(default)s]', default=5, type=float)

    ########## Workflow ##########
    master = Controller(ap)
    
    wf = Workflow(master)
    
    # If only 1 snakemake command should be run
    if master.snake:
        logging.info('Only running '+master.snake+' snakefile')
        wf.run(snakefile=globals()['WORKFLOW_'+master.snake])
    # If not, run the actual workflow
    else:
   
        # Preprocess 
        logging.info('Filtering bins')
        wf.run(snakefile=WORKFLOW_FILTER)

        # Annotation and gene clustering
        logging.info('Classifying genomes with GTDB-tk')
        wf.run(snakefile=WORKFLOW_GTDBTK)

        logging.info('Clustering genes and parsing GTDB-tk results')
        wf.run(snakefile=WORKFLOW_PARSE_GTDBTK)

        logging.info('Filtering of the gene clusters and readmapping')
        wf.run(snakefile=WORKFLOW_FILTER_GENE_CLUSTERS)

        # Signature genes
        logging.info('Identifying signature genes')
        logging.debug('Creating a gene count matrix of the readmappings')
        wf.run(snakefile=WORKFLOW_GENE_COUNT_MAT)
        logging.debug('Sorting the matrix to only contain genes, that do not cluster across the Metagenomic Species')
        wf.run(snakefile=WORKFLOW_PRESCREENING_GENES)
        logging.debug('Identifying the signature genes')
        wf.run(snakefile=WORKFLOW_SIGNATURE_GENES)

        # Phylogenies
        logging.info('Inferring MGS phylogenies')
        logging.debug('Identifying outgroups and marker genes for MGS phylogenies')
        wf.run(snakefile=WORKFLOW_OUTGROUP)
        logging.debug('Pileup of signature genes for calling SNVs')
        wf.run(snakefile=WORKFLOW_PILEUP)
        logging.debug('Parsing pileup data')
        wf.run(snakefile=WORKFLOW_PILEUP_PARSE)
        logging.debug('Generaring alignments')
        wf.run(snakefile=WORKFLOW_ALIGNMENT)
        logging.debug('Generaring phylogenies')
        wf.run(snakefile=WORKFLOW_PHYLO)

        # Gene vs tax
        logging.info('Inferring taxonomic scope of genes')
        wf.run(snakefile=WORKFLOW_GENE_TAX)

if __name__ == '__main__':
    cli()



#!/usr/bin/env python3

import re
import sys
import itertools

import pandas as pd
import numpy as np

####### Read #######
tax_df = pd.read_csv(snakemake.input['tax'], header=None, names=('MGS', 'Domain', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species'), sep='\t')
mgs_df = pd.read_csv(snakemake.input['mgs'], header=None, names=('MGS', 'Taxonomy', 'Clusters'), sep='\t')
clust_df = pd.read_csv(snakemake.input['cluster'], header=None, names=('GeneCluster', 'Gene'), sep='\t')
vamb_df = pd.read_csv(snakemake.input['vamb'], header=None, names=('Bin', 'Contig'), sep='\t')

cutoff = float(snakemake.params['cutoff'])

###### Transform ######
# Expand mgs data.frame to contain a row per cluster for easy merging
mgs_zip = zip(list(mgs_df['MGS']), list(mgs_df['Clusters']))
mgs_list = list(itertools.chain.from_iterable([[('Cluster'+str(x[0]), y) for y in str(x[1]).split(',')] for x in mgs_zip]))
mgs_df_exp = pd.DataFrame(mgs_list, columns=('MGS', 'Cluster'))

####### Merging ######
# Merge taxonomy
tax_df_new = tax_df.merge(mgs_df_exp, on='MGS', how='left')

# Fill cluster info for NAs
tax_df_new.loc[tax_df_new.Cluster.isna(), 'Cluster'] = [re.sub('Cluster', '', x) for x in list(tax_df_new.loc[tax_df_new.Cluster.isna(), 'MGS'])]

# Add contig column to gene clustering
clust_df['Contig'] = [re.sub('_[0-9]*$', '', x) for x in clust_df['Gene']]

# First merge
merge_df1 = clust_df.merge(vamb_df, on='Contig')

# Add cluster column to bin table
merge_df1['Cluster'] = [re.sub('.*_', '', x) for x in merge_df1['Bin']]

# Second merge
merge_df2 = merge_df1.merge(tax_df_new, on='Cluster')

# Write first table
merge_df1[['GeneCluster', 'Bin']].to_csv(snakemake.output[1], sep='\t', index=False)

####### Summarising ######
# Split
merge_single = merge_df2.drop_duplicates(subset='GeneCluster', keep=False)
merge_dup = merge_df2.loc[merge_df2.duplicated(subset='GeneCluster', keep=False), :]

count_df = pd.DataFrame(merge_dup.groupby('GeneCluster')[['Domain', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species']].value_counts(normalize=True,dropna=False))
count_df.reset_index(inplace=True)

# All those above cutoff are consistent at Species level
species_df = count_df[(count_df[0] > cutoff) & ~(count_df['Species'].isna())]

# Traverse through taxonomy, aggregate and check if cutoff is met
# Genus level
tmp_df = count_df[~count_df['GeneCluster'].isin(species_df['GeneCluster'])].groupby(['GeneCluster', 'Domain', 'Phylum', 'Class', 'Order', 'Family', 'Genus'], dropna=False)[0].sum().reset_index()
genus_df = tmp_df[(tmp_df[0] > cutoff) & ~(tmp_df['Genus'].isna())]

# Family level
tmp_df = tmp_df[~tmp_df['GeneCluster'].isin(genus_df['GeneCluster'])].groupby(['GeneCluster', 'Domain', 'Phylum', 'Class', 'Order', 'Family'], dropna=False)[0].sum().reset_index()
family_df = tmp_df[(tmp_df[0] > cutoff) & ~(tmp_df['Family'].isna())]

# Order level
tmp_df = tmp_df[~tmp_df['GeneCluster'].isin(family_df['GeneCluster'])].groupby(['GeneCluster', 'Domain', 'Phylum', 'Class', 'Order'], dropna=False)[0].sum().reset_index()
order_df = tmp_df[(tmp_df[0] > cutoff) & ~(tmp_df['Order'].isna())]

# Class level
tmp_df = tmp_df[~tmp_df['GeneCluster'].isin(order_df['GeneCluster'])].groupby(['GeneCluster', 'Domain', 'Phylum', 'Class'], dropna=False)[0].sum().reset_index()
class_df = tmp_df[(tmp_df[0] > cutoff) & ~(tmp_df['Class'].isna())]

# Phylum level
tmp_df = tmp_df[~tmp_df['GeneCluster'].isin(class_df['GeneCluster'])].groupby(['GeneCluster', 'Domain', 'Phylum'], dropna=False)[0].sum().reset_index()
phylum_df = tmp_df[(tmp_df[0] > cutoff) & ~(tmp_df['Phylum'].isna())]

# Domain level
tmp_df = tmp_df[~tmp_df['GeneCluster'].isin(phylum_df['GeneCluster'])].groupby(['GeneCluster', 'Domain'], dropna=False)[0].sum().reset_index()
domain_df = tmp_df[(tmp_df[0] > cutoff) & ~(tmp_df['Domain'].isna())]

# No level
tmp_df = tmp_df[~tmp_df['GeneCluster'].isin(domain_df['GeneCluster'])].groupby(['GeneCluster'])[0].sum().reset_index()

summary_df = pd.concat([species_df, genus_df, family_df, order_df, class_df, phylum_df, domain_df, tmp_df])
del summary_df[0]

summary_df.to_csv(snakemake.output[0], sep='\t', index=False, na_rep='NA')




#!/usr/bin/env python3

import re
import sys

import pandas as pd
import numpy as np

from scipy.sparse import lil_matrix
from Bio import SeqIO

# Get clustering table
dat = pd.read_csv(snakemake.input['cluster'], header=None, names=('Cluster', 'Gene'), sep='\t')

# Get cluster index
cluster_index = list(set(dat['Cluster']))
n_clust = len(cluster_index)

# Turn into dicts
clust_gene = dict(zip(dat['Gene'], dat['Cluster']))
clust_index = dict(zip(cluster_index, range(n_clust)))

# Init graph as sparse array
G = lil_matrix((n_clust, n_clust))

# Read genes and save in graph
k = 0
prev_prog = 0
prev_contig = None
prev_cluster = None
prev_strand = None
n_gene = dat.shape[0]
with open(snakemake.input['faa']) as fh:
    for gene in SeqIO.parse(fh, 'fasta'):
        # Get contig name
        contig = re.sub('_[0-9]*$', '', str(gene.id))
        
        # Get gene cluster
        cluster = clust_gene[str(gene.id)]
       
        # Get strand
        strand = str(gene.description).split()[6]

        # If we are on the same contig and strand as previous gene, genes are adjacent
        if contig == prev_contig and strand == prev_strand:
            # Get indices
            i = clust_index[cluster]
            j = clust_index[prev_cluster]
            G[i, j] += 1
            G[j, i] += 1

        # Save previous cluster
        prev_cluster = cluster

        # Save previous contig
        prev_contig = contig

        # Save previous strand
        prev_strand = strand

        prog = round(100*k/n_gene)
        if prog > prev_prog:
            print(str(prog)+'%')
        prev_prog = prog
        k += 1
      
# Write graph
with open(snakemake.output['graph'], 'w') as fh:
    print(G, file = fh)

# Write cluster indices
with open(snakemake.output['index'], 'w') as fh:
    for k in zip(range(n_clust), cluster_index):
        fh.write('{} {}\n'.format(k[0], k[1]))



#!/usr/bin/env python3

import glob
import re
import sys
import os

import multiprocess as mp

from Bio import SeqIO

# Constants
align_dir = snakemake.params['ali_dir']
samp_tab = snakemake.input[0]
out_dir = snakemake.output[0]
threads = snakemake.resources['cores']

# Make out dir
try:
    os.mkdir(out_dir)
except FileExistsError:
    pass

# Get list of clusters
clusters = [x for x in os.listdir(align_dir)]

# Get list of samples
samples = {}
samples['Outgroup'] = ''
with open(samp_tab, 'r') as handle:
    for x in handle.readlines():
        samples[x.strip().split(',')[0]] = ''

# For each cluster:
'''
Load sequences
Concat
Write to a single fasta file
Write partition file for iqtree 
'''

def concat(clust, samples = samples):

    # Wrapper to write partition file
    with open(os.path.join(out_dir, clust+'.part'), 'w') as part_handle:

        # Init
        gene_start = 1
        seq_len = 1

        # Read through alignments and fill in
        for afile in glob.glob(os.path.join(align_dir, clust, clust+'_'+'*.fna')):
            seqs = {}

            # Load in seqs
            with open(afile, 'r') as handle:
                for fa in SeqIO.parse(handle, 'fasta'):
                    seqs[fa.id] = str(fa.seq)
                    seq_len = len(fa.seq)
           
            # Loop through samples and fill in
            for samp in samples:
                if samp in seqs.keys():
                    samples[samp] += seqs[samp]
                else:
                    samples[samp] += '-' * seq_len

            # Write partition file
            gene_name = re.sub('.*'+clust+'_(.*)\.fna$', r'\1', afile)
            gene_end = gene_start + seq_len - 1
            part_handle.write('DNA, '+str(gene_name)+' = '+str(gene_start)+'-'+str(gene_end)+'\n')
            gene_start += seq_len

    # Write fasta
    samples = {k:v for (k,v) in samples.items() if len(v.replace('-', '')) > 0}

    with open(os.path.join(out_dir, clust+'.fna'), 'w') as handle:
        for samp in samples:
            handle.write('>'+samp+'\n'+samples[samp]+'\n')

pool = mp.Pool(threads)
concats = list(pool.imap(concat, set(clusters)))




import sys
import yaml
import re
import os
import subprocess
import logging
import glob

class Message(object):

    def __init__(self, obj):
        for key, val in vars(obj).items():
            setattr(self, key, val)
        
        self.commands = []

    def add(self, x):
        
        self.commands.append(x)

    def check(self):
        
        # Get latest command
        last = self.commands[-1]

        if 'workflow/filter.Snakefile' in last:
            n_clust= 0
            n_bins = 0
            for path, dirs, files in os.walk(os.path.join(self.output, 'clusters')):
                for d in dirs:
                    n_clust += 1
                for f in files:
                    n_bins += 1
            logging.info(str(n_bins) + ' bins in ' + str(n_clust) + ' VAMB clusters left after filtering')

        if 'workflow/gtdbtk.Snakefile' in last:
            n_clust = len(glob.glob(self.output+'gtdbtk/*/classify/'))
            logging.info(str(n_clust) + ' clusters could be classified')

        if 'workflow/parse_gtdbtk.Snakefile' in last:
            with open(os.path.join(self.output, 'tabs', 'metagenomicspecies.tab'), 'r') as fh:
                n_clust = 0
                n_mgs = 0
                for line in fh:
                    n_clust += len(line.strip().split('\t')[2].split(','))
                    n_mgs += 1
            logging.info(str(n_clust) + ' VAMB clusters merged into ' + str(n_mgs) + ' metagenomic species')

        if 'workflow/filter_geneclusters.Snakefile' in last:
            logging.debug('Readmapping to the clustered genes has been done.')

        if 'workflow/gene_count_mat.Snakefile' in last:
            logging.debug('A gene count matrix has been created - summarizing the readmappings for all genes in all samples')

        if 'workflow/prescreening_genes.Snakefile' in last:
            total_clust=len(glob.glob(self.output+'signature_genes/clusters/*'))
            logging.info('A total of ' + str(total_clust) + ' clusters are included in the analysis.')            

        if 'workflow/signature_genes.Snakefile' in last:
            no_SG = 0
            for f in os.listdir(os.path.join(self.output, 'signature_genes/screened/')):
                path = os.path.join(self.output, 'signature_genes/screened/', f)
                if os.path.isfile(path):
                    if (os.path.getsize(path)<60):
                        no_SG += 1
            logging.info('Too few genes was present in ' + str(no_SG) + ' MGSs in order to identify Signature Genes. The relative abundance has been set to 0.')

        if 'workflow/phylo.Snakefile' in last:
            k = 0
            for f in glob.glob(os.path.join(self.output, 'phylo', 'trees', '*.nwk')):
                if os.path.getsize(f) > 0:
                    k += 1
            logging.info('Phylogenies generated for '+str(k)+' metagenomic species')
            



#!/usr/bin/env python3

import sys
import re

import pandas as pd

from collections import Counter
from itertools import chain

'''
Worflow of this script, after reading and preparing input data:
1. Finding possible roots for each cluster, from the taxonomy information (Optimal: same family, different genus))
2. Find common GTDB-tk marker genes for each cluster 
3.1. Filter the possible roots to only those sharing marker genes with the cluster
3.2. Pick root for each cluster (Most shared markers)
4.1 Find the geneIDs for the roots (redundant gene catalog)
4.2 Find the geneIDs for the marker genes for clusters (non-redundant gene catalog)
'''

#######################################################################################################################
# Constants
marker_gene_cluster_prevalence = float(snakemake.params.marker_gene_cluster_prevalence)
min_gtdb_markers = int(snakemake.params.min_gtdb_markes)
output_file = snakemake.output[0]

#######################################################################################################################
# Exceptions
class MarkerGeneError(Exception):
    pass

#######################################################################################################################
# Read data
tax = pd.read_csv(snakemake.input[0], sep='\t', header=None, names=('user_genome', 'classification'))

tax_cl = pd.read_csv(snakemake.input[1], sep='\t', header=None, names=('cluster', 'classification', 'clusters'))

try:
    unq_bac = pd.read_csv(snakemake.input[2], sep='\t', header=None)
    unq_bac = unq_bac[unq_bac[0].isin([y for x,y in zip(tax['classification'],tax['user_genome']) if 'd__Bacteria' in x])]
except Exception:
    unq_bac = pd.DataFrame()

try:
    unq_ar = pd.read_csv(snakemake.input[3], sep='\t', header=None)
    unq_ar = unq_ar[unq_ar[0].isin([y for x,y in zip(tax['classification'],tax['user_genome']) if 'd__Archaea' in x])]
except Exception:
    unq_ar = pd.DataFrame()

gtdb_markers_bin = pd.read_csv(snakemake.input[4], sep=r'\s+', header=None, names=('Gene','Marker','Bin','GeneCluster'), index_col=False)

gene_clusters_nr = pd.read_csv(snakemake.input[5], sep='\t', header=None, names=('Name','Length', 'Offset', 'Linebases', 'Linewidth'))

#######################################################################################################################
# Prepare data

# Make non-redundant gtdb_markers_bin file
gtdb_markers_bin_nr = gtdb_markers_bin[gtdb_markers_bin['GeneCluster'].isin(gene_clusters_nr['Name'])]

# Combine markers
unq_marker_df = pd.concat([unq_bac, unq_ar], axis=0)

# Get unique marker genes
unq_marker_dict = unq_marker_df.set_index(0)[1].to_dict()
unq_marker_dict = {k: set(v.split(',')) for (k, v) in unq_marker_dict.items()}

# Bin cluster matching
bin_clust_dict_inv = {k: v for k, v in zip(unq_marker_dict.keys(), [re.search(r'_([0-9]*)$', x).group(1) for x in unq_marker_dict.keys()])}

# Inverse above dict for better lookup
bin_clust_dict = {}
for k, v in bin_clust_dict_inv.items():
    bin_clust_dict[v] = bin_clust_dict.get(v, []) + [k]

# Split taxonomy to columns
tax_split = pd.DataFrame([x.split(';') for x in tax['classification']])

tax = pd.concat([tax.reset_index(drop=True), tax_split.reset_index(drop=True)], axis=1)

tax_clust = pd.DataFrame([x.split(';') for x in tax_cl['classification']])
tax_clust['cluster'] = tax_cl['cluster']

#######################################################################################################################
# For each cluster find possible roots
# also save domain data for later
'''
root_dict is a dict with the cluster number as key
and a nested list as value
with the first level in the list being the different taxonomic levels, family, order, class, phylum, domain
and the second level containing names of bins, which are possible roots

E.g. Assuming entire taxonomy is known:
Bins in root_dict[clust][0] have similar family as cluster, but different genus
Bins in root_dict[clust][1] have similar order as cluster, but different family
'''
root_dict = {}
domain_dict = {}
for row in tax_clust.iterrows():
    # Start level is family
    level = 4
    # Make sure taxonomy is known
    # Work up taxonomy if not known
    unknown_tax = True
    
    while unknown_tax:
        if row[1][level] is None:
            level -= 1
            if level < 0:
                sys.exit('Taxonomy unknown')
        else:
            unknown_tax = False

    # Find clusters similar at the level, but different at the more specific level (genus first iter)
    root_list = list()
    while level > -1:
        this = list(tax[(tax[level] == row[1][level]) & (tax[level+1] != row[1][level+1])]['user_genome'])
        root_list.append(this)
        level -= 1
    
    root_dict[row[1]['cluster']] = root_list
    domain_dict[row[1]['cluster']] = row[1][0]


#######################################################################################################################
# Marker genes per cluster
'''
marker_genes_dict is a dict with the cluster number as key
and a set as value
which contains the marker genes present in a fraction higher than "marker_gene_cluster_prevalence" in the cluster
'''
marker_genes_dict = {}
for clust in set(tax_cl['cluster']):

    # List of bins in cluster
    bin_list = bin_clust_dict[str(clust)] 
   
    # Get markers for each bin in cluster
    marker_list = [unq_marker_dict[x] for x in bin_list]
  
    # Only non-redundant markers for the cluster
    bin_marker_nr = gtdb_markers_bin_nr[gtdb_markers_bin_nr['Bin'].isin(bin_list)]
    marker_list = [{x for x in y if x in list(bin_marker_nr['Marker'])} for y in marker_list] 

    # Unpack and count
    marker_list_count = Counter(list(chain.from_iterable(marker_list)))

    # Get those above prevalence threshold
    marker_list_prev = [k for k, v in marker_list_count.items() if v/len(bin_list) >= marker_gene_cluster_prevalence]

    if len(marker_list_prev) == 0:
        sys.exit('No shared marker genes above prevalence threshold for cluster '+str(clust))

    marker_genes_dict[clust] = set(marker_list_prev)

#######################################################################################################################
# Find roots that are sharing marker genes
'''
final_roots is a dict with the cluster number as key
and a tuple as value
with first part containing name of bin, which should be the root (most overlapping marker genes, in most specific tax level (except genus))
and second part containing the names of the marker genes shared by the cluster and the root
'''
final_roots = dict()
for clust in set(tax_cl['cluster']):
    
    # Marker genes of focal cluster
    marker_gene_focal = marker_genes_dict[clust]

    # Marker genes of possible roots
    marker_gene_roots = [[unq_marker_dict[x] for x in level] for level in root_dict[clust]]
    bin_name_roots = [[x for x in level] for level in root_dict[clust]]

    # Get only overlapping marker genes
    marker_gene_overlap = [[list(x.intersection(marker_gene_focal)) for x in level] for level in marker_gene_roots]
  
    # At least X marker genes or max possible from focal
    min_markers = min(min_gtdb_markers, len(marker_gene_focal))

    # Only roots with at least "min_markers" overlapping genes
    marker_gene_overlap = [[(y, x) for x,y in zip(level1,level2) if len(x)>=min_markers] for level1,level2 in zip(marker_gene_overlap,bin_name_roots)]

    # Traverse up taxonomy to find the most specific root
    for level in marker_gene_overlap:
        # If no overlap check next level
        if len(level) == 0:
            continue
        # If overlap, pick the most overlapping
        else:
            ll_tmp = [len(x[1]) for x in level]
            root_bin = level[ll_tmp.index(max(ll_tmp))]        
            break
    
    final_roots[clust] = root_bin

#######################################################################################################################
# Find the geneIDs for the marker genes for the clusters and roots (bins, redundant gene catalog)
df_list = []
for clust in final_roots:

    root = final_roots[clust][0]
    markers = final_roots[clust][1]
   
    # Subset
    clust_markers = gtdb_markers_bin[(gtdb_markers_bin['Bin'].isin(bin_clust_dict[str(clust)])) & (gtdb_markers_bin['Marker'].isin(markers))]     
    bin_markers = gtdb_markers_bin[(gtdb_markers_bin['Bin'] == root) & (gtdb_markers_bin['Marker'].isin(markers))]     

    # Get most common Gene cluster and Marker gene combinations
    count_markers = Counter(list(zip(clust_markers['Marker'], clust_markers['GeneCluster'])))
    shared_markers = count_markers.most_common()
    shared_markers_df = pd.DataFrame([x[0] for x in shared_markers], columns = ['Marker','GeneCluster'])

    # Try removing all duplicates and check if enough markers are present
    try:
        shared_markers_df_copy = shared_markers_df.copy()
        shared_markers_df_copy.drop_duplicates(subset=['GeneCluster'], keep=False, inplace=True)
        shared_markers_df_copy.drop_duplicates(subset=['Marker'], keep=False, inplace=True)
        
        if len(shared_markers_df_copy) < min_gtdb_markers:
            raise MarkerGeneError('Fewer than '+str(min_gtdb_markers)+' marker genes uniquely matching gene clusters for cluster '+str(clust))
    
    # If not possible, keep first if there are duplicates
    # This means some marker genes can be represented by multiple gene clusters
    # and that gene clusters can match to multiple marker genes
    # In case of duplicates the first, most common geneCluster-markerGene combination, is chosen
    except MarkerGeneError:
        shared_markers_df_copy = shared_markers_df.copy()
        shared_markers_df_copy.drop_duplicates(subset=['GeneCluster'], inplace=True)
        shared_markers_df_copy.drop_duplicates(subset=['Marker'], inplace=True)
            
    # Subset 
    min_markers = min(min_gtdb_markers, len(shared_markers_df_copy))
    shared_markers_sub = shared_markers_df_copy.copy().iloc[:min_markers, ]
    
    # If a root bin has two of the same marker, keep one (pseudo-random) 
    bin_markers = bin_markers.drop_duplicates(subset=['Marker']).copy()
    
    # Ensure that cluster and bin have the same markers
    bin_markers = bin_markers[bin_markers['Marker'].isin(shared_markers_sub['Marker'])]
    shared_markers_sub = shared_markers_sub[shared_markers_sub['Marker'].isin(bin_markers['Marker'])]

    # Order
    shared_markers_sub.sort_values(by='Marker', inplace=True)
    bin_markers.sort_values(by='Marker', inplace=True)
    
    # Just a sanity check
    if list(shared_markers_sub['Marker']) != list(bin_markers['Marker']):
        print('Cluster: '+str(clust))
        print(shared_markers_sub)
        print(bin_markers)
        sys.exit('Oops, cluster markers and root markers do not match up')
   
    out_df = pd.DataFrame({'Cluster': len(bin_markers)*[clust],
                           'Root': len(bin_markers)*[root],
                           'Marker': list(bin_markers['Marker']),
                           'GeneCluster': list(shared_markers_sub['GeneCluster']),
                           'GeneRoot': list(bin_markers['Gene'])})
    
    df_list.append(out_df)

out_df_concat = pd.concat(df_list)
out_df_concat.to_csv(output_file, sep='\t', index=False, header=False)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created 2022

@author: trinezachariasen
"""

# importing the species collections and it's corresponding MGS-ids included in the Metagenomic Species Collection
import csv
import sys
csv.field_size_limit(sys.maxsize) # I have large entries in my csv-file with gene-id's

cluster_MGS_d = {}
with open(snakemake.input[0] , 'r') as csv_file:
    for row in csv.reader(csv_file, delimiter='\t'):
        for cluster in row[2].split(','):
            cluster_MGS_d[cluster] = row[0]

d = {}
with open(snakemake.input[0], 'r') as csv_file:
    for row in csv.reader(csv_file, delimiter='\t'):
            d[row[0]] = row[2] # setting the MGS as key and it's included MGS-ids as value

# importing the VAMB clustefile as dict with the geneid as key and it's corresonding MGS as value
id_dict = {}
with open(snakemake.input[2], 'r') as id_file:
    for row in csv.reader(id_file, delimiter='\t'):
        cluster = row[0].split("_")[-1]
        id_dict[row[1]] = cluster

## Creating a file only containing the representative / nonredundant genes
## Based on the clustering file from mmseqs
gene_file = snakemake.input[1]
outfile = snakemake.output[0]


excluded_counter = set()
cluster_counter = set()
contig_gene = {}

with open(gene_file) as f:
    for line in f.readlines():
        [raw_rep, raw_gene] = line.rstrip().split('\t') # the geneIDs
        rep = "_".join(raw_rep.split("_")[:-1]) # representative contigID
        gene = "_".join(raw_gene.split("_")[:-1]) # clustered contigID

        repMGS = id_dict[rep] # mgsID of representative
        geneMGS = id_dict[gene] # mgsID of the gene
        if (repMGS not in cluster_MGS_d):
            cluster_MGS_d[repMGS]=repMGS

        if (rep==gene): #if the representative gene is same as clustered we keep it
            cluster_counter.add(raw_rep)
            contig_gene[raw_rep] = cluster_MGS_d[repMGS]
        try:
            if ((d[geneMGS].find(repMGS)==0)): # if the MGS of the gene is included in the representative gene Metagenomic Species Collection
                cluster_counter.add(raw_rep)
                contig_gene[raw_rep] = cluster_MGS_d[repMGS]
        except KeyError:
            pass

        if (geneMGS==repMGS):  # if the MGS of the Gene is same as representative gene MGS
             cluster_counter.add(raw_rep)
             contig_gene[raw_rep] = cluster_MGS_d[repMGS]
        else: # if the representative MGSid is not equal to the MGS of the gene it represents
            excluded_counter.add(raw_rep) 
            contig_gene.pop(raw_rep, None)
             
# identifying the representative genes, which do not cover more than one cluster
rep_to_keep = list(cluster_counter.difference(excluded_counter))

with open(outfile, "w") as out:
    out.write("\n".join(rep_to_keep))

# sorting the geneID_clusterID to become geneID_collectionID
with open(snakemake.output[1],'w') as f:
    w = csv.writer(f, delimiter='\t')
    for gene in rep_to_keep:
       w.writerow([gene, contig_gene[gene]])




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 13 09:59:31 2022

@author: trinezachariasen
"""

import csv
import sys
csv.field_size_limit(sys.maxsize) # I have large entries in my csv-file with gene-id's


# importing set with the genes we want to keep
gene_set = set()
with open(snakemake.input[1], 'r') as id_file:
    for row in csv.reader(id_file):
        gene_set.add(row[0])
    
out_mat = open(snakemake.output[0],"w")

# Going through the gene count matrix (containing all genes and all clusters, sorting away the genes, that span multiple clusters)
with open(snakemake.input[0], 'r') as mat_file:
    out_mat.write(mat_file.readline()) # writing the header to the outfile
    
    for row in csv.reader(mat_file, delimiter='\t'): # looping through the genes
#        gene = "_".join(row[0].split("_")[:-1])
        if (row[0] in gene_set): #if the gene is found in the gene_set we want to keep, write it to the smaller matrix-file
            out_mat.write("\t".join(row) + "\n")




import os
import logging
import sys
import pkg_resources
import re

class Controller(object):

    def __init__(self, ap):
     
        args = ap.parse_args()

        for k,v in args.__dict__.items():
            setattr(self, k, v)
        
        # Logger
        logging.basicConfig(format='\033[36m'+'[%(asctime)s] %(levelname)s:'+'\033[0m'+' %(message)s', datefmt='%Y-%m-%d %H:%M:%S', level=self.log_lvl)
        logging.info('Running MAGinator version {}'.format(pkg_resources.require("maginator")[0].version))

        # Force consistency
        self.output = os.path.join(self.output, '')

        # Check cluster info input
        if self.cluster is not None:
            if self.cluster_info is not None:
                if len(re.findall('{cores}|{mem_gb}|{runtime}', self.cluster_info)) != 3 or len(re.findall('{.*?}', self.cluster_info)) != 3: 
                    logging.error('cluster_info has to contain the following special strings: {cores}, {mem_gb}, and {runtime}')
                    sys.exit()
                else:
                    tmp_info = re.sub('{cores}|{mem_gb}|{runtime}','',self.cluster_info)
                    if not bool(re.match('^[a-zA-Z0-9-_ =:,.]+$', tmp_info)):
                        logging.error('Invalid characters in cluster_info')
                        sys.exit()
            else:
                logging.error('cluster_info is required when running on a compute cluster')
                sys.exit()

        # Check input and output
        self.check_params()
        self.check_out()
        if not any([self.unlock, self.only_conda, self.snake is not None]):
            self.check_vamb()
            self.check_reads()
            self.check_contigs()
        self.write_params(args)

    def check_params(self):
        '''
        Return error if parameters are not allowed
        '''

        if self.tax_scope_threshold < 0.5:
            logging.error('tax_scope_threshold lower than 0.5 can lead to unexpected results')
            sys.exit()
        if self.min_af < 0.5:
            logging.error('min_af lower than 0.5 can lead to unexpected results')
            sys.exit()

    def check_out(self):
        '''
        Creates output directory unless it exists already
        '''
        try:
            os.makedirs(self.output+'logs')
        except FileExistsError:
            logging.warning('Output directory '+self.output+' already exists')

    def check_vamb(self):
        '''
        Check the vamb input file
        '''
        logging.debug('Checking VAMB input')
        with open(self.vamb_clusters, 'r') as fh:
            vamb_file = [x.strip().split() for x in fh.readlines()]
        
        # Extract info
        self.bin_set = set([x[0] for x in vamb_file])
        self.contig_set = set([x[1] for x in vamb_file])
        self.sample_set = set([re.sub('_[0-9]+$', '', x) for x in self.bin_set])
        
        try:
            self.cluster_set = set([int(re.sub('.*_', '', x)) for x in self.bin_set])
        except Exception:
            logging.error('Cluster names should be numeric. 1st column in the VAMB input file should be of the form SampleName_ClusterName')
            sys.exit()

        if not all([len(x)==2 for x in vamb_file]):
            logging.error('VAMB clusters.tsv is invalid. All lines should contain 2 tab-separated columns')
            sys.exit()            

        del self.bin_set
        del vamb_file

    def check_reads(self):
        '''
        Check that the reads file is of the correct format
        Check if read files exists
        '''
        logging.debug('Checking read info file')
        
        with open(self.reads, 'r') as fh:
            read_file = [x.strip().split(',') for x in fh.readlines()]

        if not all([len(x)==3 for x in read_file]):
            logging.error('read file is invalid. All lines should contain 3 comma-separated columns')
            sys.exit()            
        
        # Check if samples names match
        sample_set = set([x[0] for x in read_file])
        if not self.sample_set == sample_set:
            logging.error('Sample names in read file do not match those in the VAMB clusters.tsv file')
            sys.exit()
       
        # Check if read files exists
        if not all([os.path.isfile(y) for y in [x[1] for x in read_file]]):
            logging.error('Cannot find fastq files the forward reads')
            sys.exit()
        if not all([os.path.isfile(y) for y in [x[2] for x in read_file]]):
            logging.error('Cannot find fastq files the reverse reads')
            sys.exit()
            
    def check_contigs(self):
        '''
        Load contig file and check that it matches the clusters.tsv file
        '''
        logging.debug('Checking contig fasta file')
        
        contig_names = list()
        with open(self.contigs, 'r') as fh:
            for line in fh:
                if line.startswith('>'):
                    contig_names.append(line.strip().split()[0][1:])

        if not set(contig_names) == self.contig_set:
            logging.error('Contig names in fasta file do not match contig names in VAMB clusters.tsv file')
            sys.exit()

        del self.contig_set
        del contig_names

    def write_params(self, args):
        '''
        Write parameters for snakemake workflows to a file
        '''

        pars = vars(args)
        self.params = self.output+'parameters.tab'
        fh = open(self.params, 'w')
        for k, v in pars.items():
            fh.write('{}\t{}\n'.format(k, v))
        fh.close()



#!/usr/bin/env python3

import sys
import re
import os
import shutil
import copy
import pickle

import multiprocess as mp
import pandas as pd

from Bio import SeqIO 
from collections import defaultdict

'''
Workflow of this script:
1. Get lengths of genes
2. Read mpileup files
3. Convert to a dict of genes with dicts of samples inside
4. Split to dict cluster-wise
5. For each cluster:
5.1. Get signature genes and marker genes for cluster
5.2. Get marker genes for root
5.3. Collect data, such as total non-N characters, N genes mapped per sample, coverage per sample, allele frequency threshold exceeded per sample
5.4. Filter samples per cutoffs, see above
5.5. Write statistics to file
5.6. Write sequences to fasta files for alignment
'''


#######################################################################################################################
# Constants
const_min_af = float(snakemake.params.min_af)
const_min_depth = float(snakemake.params.min_depth)
threads = int(snakemake.resources.cores)
min_nonN = float(snakemake.params.min_nonN)
min_marker_genes = int(snakemake.params.min_marker_genes)
min_signature_genes = int(snakemake.params.min_signature_genes)
pileup_dir = snakemake.input.pileup
ref_fasta = snakemake.input.ref
root_fasta = snakemake.input.root
sig_gene_tab = snakemake.input.sig
mark_gene_tab = snakemake.input.mark
tax_tab = snakemake.input.tax
align_out = snakemake.output.align
stat_out = snakemake.output.stat
stat_gene_out = snakemake.output.gene 
pileup_dump = snakemake.output.dump

#######################################################################################################################
try:
    os.mkdir(align_out)
except Exception:
    pass

def main():
    '''
    The workflow
    '''

    # Get gene lengths
    gene_len_dict = read_refs(ref_fasta)
    
    # Get root marker genes
    root_markers = read_roots(root_fasta)

    # Sample names
    mp_files = os.listdir(pileup_dir)
    sample_names = [re.sub('^(.*).mp$', '\\1', x) for x in mp_files]

    # Check if pileup files have been read before
    if os.path.exists(pileup_dump):
        with open(pileup_dump, "rb") as fhi:
            pileup_dict_gene = pickle.load(fhi)
    else:
        # Pileup files are read
        pool = mp.Pool(threads)
        pileups = list(pool.imap(mpileup_read, [os.path.join(pileup_dir, x) for x in mp_files]))

        # Change from list of samples with dicts of genes to dict of genes with dict of samples
        pileup_dict_samp = {k: v for k, v in zip(sample_names, pileups)}
        pileup_dict_gene = inverse_dict(pileup_dict_samp)

        # Dump file for re-running
        with open(pileup_dump, "wb") as fho:
            pickle.dump(pileup_dict_gene, fho)

    # Gene info
    signature_genes = pd.read_csv(sig_gene_tab, sep='\t', header=None, names=('Gene','Cluster'))
    signature_genes.dropna(inplace=True)
    tax_matrix = pd.read_csv(tax_tab, sep='\t', header=None)
    tax_matrix.dropna(inplace=True)
    signature_genes = signature_genes[signature_genes['Cluster'].isin(tax_matrix[0])]
    marker_genes = pd.read_csv(mark_gene_tab, sep='\t', header=None, names=('ClusterNo','Root','Marker','ClusterGene','RootGene'))
    marker_genes['Cluster'] = ['Cluster'+str(x) for x in marker_genes['ClusterNo']]

    with open(stat_out, 'w') as fh:
        fh.write('Cluster\tSample\tNonN\tSignature_N\tSignature_Cov\tSignature_AF\tMarker_N\tMarker_Cov\tMarker_AF\n')
    with open(stat_gene_out, 'w') as fh:
        fh.write('Cluster\tSample\tGene\tNonN\tSignature_Cov\tMarker_Cov\tSignature_AF\tMarker_AF\n')
    
    for clust in set(signature_genes['Cluster']):

        print(clust)

        # Get genes for specific cluster    
        marker_gene_list = list(marker_genes[marker_genes['Cluster'] == clust]['ClusterGene'])
        signature_gene_list = list(signature_genes[signature_genes['Cluster'] == clust]['Gene'])

        # If the signature gene is also a marker gene, only keep it as a marker gene
        signature_gene_list = [x for x in signature_gene_list if x not in marker_gene_list]

        # Extact genes for specific cluster and put in new dicts
        marker_dict = {x: pileup_dict_gene.get(x) for x in marker_gene_list if pileup_dict_gene.get(x) is not None}
        signature_dict = {x: pileup_dict_gene.get(x) for x in signature_gene_list if pileup_dict_gene.get(x) is not None}
   
        # Get marker genes for root
        root_gene_list = list(marker_genes[marker_genes['Cluster'] == clust]['RootGene'])
        root_gene_seqs = [root_markers[x] for x in root_gene_list]

        # Remove the few genes not found in pileup (bam)
        marker_gene_list = [x for x in marker_gene_list if x in marker_dict.keys()]
        signature_gene_list = [x for x in signature_gene_list if x in signature_dict.keys()]

        # Add root marker genes to marker_dict
        for gene_no in range(len(marker_gene_list)):
            marker_dict[marker_gene_list[gene_no]]['Outgroup'] = [root_gene_seqs[gene_no], None, None, None, None, None]
        
        # Get total length of all genes for cluster
        cluster_len_best_list = [int(gene_len_dict.get(x)) for x in marker_gene_list + signature_gene_list]
        cluster_len_best = sum(cluster_len_best_list)

        # Get length of all non-N for each sample 
        cluster_len = [{k: len(v[0].replace('N','')) for k, v in pileup_dict_gene.get(x).items()} for x in marker_gene_list + signature_gene_list if pileup_dict_gene.get(x) is not None]
        cluster_len_samp = {x: sum(filter(None, [y.get(x) for y in cluster_len])) for x in sample_names}
       
        # Select on number of signature and marker genes with at least one read (present in pileup_dict_gene)
        marker_count = [[k for k,v in pileup_dict_gene.get(x).items() if len(v[0])>0] for x in marker_gene_list if pileup_dict_gene.get(x) is not None]
        marker_count_samp = {x: sum([1 for y in marker_count if x in y]) for x in sample_names}
        marker_count_too_few = set([x for x in marker_count_samp if marker_count_samp[x] < min_marker_genes])        

        signature_count = [[k for k,v in pileup_dict_gene.get(x).items() if len(v[0])>0] for x in signature_gene_list if pileup_dict_gene.get(x) is not None]
        signature_count_samp = {x: sum([1 for y in signature_count if x in y]) for x in sample_names}
        signature_count_too_few = set([x for x in signature_count_samp if signature_count_samp[x] < min_signature_genes])        

        # Remove if too many N's
        samp_remove_N = set([k for k,v in cluster_len_samp.items() if v/cluster_len_best < min_nonN])
        samp_remove_count = marker_count_too_few.union(signature_count_too_few)
        
        # Collect samples to remove
        samp_remove = samp_remove_N.union(samp_remove_count)
       
        # Coverage function
        # Sum of depth for all bases divided by sum of sequencing length
        def cov_fun(k, ll):
            ll = [ll.get(x) for x in ll if ll.get(x).get(k) is not None]
            try:
                return sum(filter(None, [y.get(k)[0] for y in ll]))/sum(filter(None, [y.get(k)[1] for y in ll]))
            except ZeroDivisionError:
                return 0

        # Get coverage data for signature genes
        signature_gene_coverage = {x: {k: (v[3],v[5]) for k, v in pileup_dict_gene.get(x).items()} for x in signature_gene_list if pileup_dict_gene.get(x) is not None}
        signature_samp_coverage = {x: cov_fun(x, signature_gene_coverage) for x in sample_names}

        # Get coverage data for marker genes
        marker_gene_coverage = {x: {k: (v[3],v[5]) for k, v in pileup_dict_gene.get(x).items()} for x in marker_gene_list if pileup_dict_gene.get(x) is not None}
        marker_samp_coverage = {x: cov_fun(x, marker_gene_coverage) for x in sample_names}
        
        # Get allele frequency data for signature genes
        signature_gene_af = {x: {k: v[4] for k, v in pileup_dict_gene.get(x).items()} for x in signature_gene_list if pileup_dict_gene.get(x) is not None}
        signature_samp_af = {x: sum(filter(None, [signature_gene_af.get(y).get(x) for y in signature_gene_af])) for x in sample_names}

        # Get allele frequency data for marker genes
        marker_gene_af = {x: {k: v[4] for k, v in pileup_dict_gene.get(x).items()} for x in marker_gene_list if pileup_dict_gene.get(x) is not None}
        marker_samp_af = {x: sum(filter(None, [marker_gene_af.get(y).get(x) for y in marker_gene_af])) for x in sample_names}
        
        # Print stats per cluster
        with open(stat_out, 'a') as fh:
            for i in sample_names:
                fh.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(str(clust),
                                str(i),
                                str(cluster_len_samp[i]/cluster_len_best),
                                str(signature_count_samp[i]),
                                str(signature_samp_coverage[i]),
                                str(signature_samp_af[i]),
                                str(marker_count_samp[i]),
                                str(marker_samp_coverage[i]),
                                str(marker_samp_af[i])))

        # Print stats per gene
        cluster_len_dict = {x: {k: len(v[0].replace('N','')) for k, v in pileup_dict_gene.get(x).items()} for x in signature_gene_list + marker_gene_list if pileup_dict_gene.get(x) is not None}
        nonn_gene = {x: {y: (int(cluster_len_dict.get(y).get(x))/int(gene_len_dict.get(y)) if cluster_len_dict.get(y).get(x) is not None else 0) for y in cluster_len_dict} for x in sample_names}
        
        with open(stat_gene_out, 'a') as fh:
            # Per sample
            for i in sample_names:
                # Per gene
                gene_sub = nonn_gene[i]
                for j in gene_sub:
                    
                    # Get multiallelic fraction
                    AF_sig = af_gene(signature_gene_af, gene_len_dict, i, j)
                    AF_mark = af_gene(marker_gene_af, gene_len_dict, i, j)
                    
                    # Get coverage info
                    sig_cov = cov_gene(signature_gene_coverage, i, j)
                    mark_cov = cov_gene(marker_gene_coverage, i, j)

                    fh.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(
                        str(clust),
                        str(i),
                        str(j),
                        str(gene_sub[j]),
                        str(sig_cov),
                        str(mark_cov),
                        str(AF_sig),
                        str(AF_mark)))


        # Delete samples
        marker_dict = delete_keys_dict(marker_dict, samp_remove)
        signature_dict = delete_keys_dict(signature_dict, samp_remove)
        
        # Remove previous alignments
        try:
            os.mkdir(os.path.join(align_out, clust))
        except FileExistsError:
            shutil.rmtree(os.path.join(align_out, clust))
            os.mkdir(os.path.join(align_out, clust))
        
        # Write to fastas
        for gene_id in marker_dict:
            sub_dct = marker_dict[gene_id]
            with open(os.path.join(align_out, clust, clust+'_'+gene_id+'.fna'), 'w') as fh:
                for fa in sub_dct:
                    if len(sub_dct[fa][0]) > 0:
                        fh.write('>'+fa+'\n'+sub_dct[fa][0]+'\n')

        for gene_id in signature_dict:
            sub_dct = signature_dict[gene_id]
            with open(os.path.join(align_out, clust, clust+'_'+gene_id+'.fna'), 'w') as fh:
                for fa in sub_dct:
                    if len(sub_dct[fa][0]) > 0:
                        fh.write('>'+fa+'\n'+sub_dct[fa][0]+'\n')

def af_gene(dd, gdd, samp, gene):
    '''
    With a nested dictionary (dd) of multiallelic counts for genes (first layer) and samples (second layer),
    and a dictionary (gdd) of gene lengths
    return the multiallelic site fraction for the gene (gene) for that sample (samp)
    '''
    try:
        AF = dd.get(gene).get(samp)
        try:
            AF = AF / int(gdd.get(gene))
        except:
            AF = 0
    except:
        AF = 'NA'
    return(AF)

def cov_gene(dd, samp, gene):
    '''
    With a nested dictionary (dd) of total depth (3) and seq_len (5) for genes (first layer) and samples (second layer) 
    return the coverage for the gene (gene) for that sample (samp)
    '''
    try:
        cov = dd.get(gene).get(samp)
        if cov is None:
            cov = 0
        else:
            try:
                cov = cov[0]/cov[1]
            except ZeroDivisionError:
                cov = 0
    except:
        cov = 'NA'
    return(cov)

def delete_keys_dict(dct, keys):
    '''
    Will delete keys (keys) from the second layer dictionaries of a deep copy of the input dictionary (dct)
    '''
    dct_copy = copy.deepcopy(dct)
    for k in keys:
        for kk in dct:
            try:
                del dct_copy[kk][k]
            except KeyError:
                pass
    return(dct_copy)

def inverse_dict(dd):
    '''
    Reverse the inner and outer dicts of a nested dictionary
    such that dd[samp][gene] becomes dd[gene][samp]
    '''
    inversed_dict = defaultdict(dict)
    for k, v in dd.items():
        for sk, sv in v.items():
            inversed_dict[sk][k] = sv

    return(dict(inversed_dict))


def parse_base_pairs(ref_base, x, min_af, min_depth):
    '''
    Parse mpileup read base column
    
    ref_base: Base of reference sequences
    x: pileup read base string
    min_af: Minimum allele frequency, if below this an N is returned

    Return:
    Str: Base
    Int: Multiallelic sites
    '''
   
    # Allele frequency indicator
    af_low = 0

    # Ignore positions of reads
    x = re.sub('\^.', '', x)
    x = x.replace('$', '')

    # Force reverse reads to forward
    x = x.upper()
    x = x.replace(',', '.')

    # Insert reference
    x = x.replace('.', ref_base)
    
    # Return N if depth too low
    if len(x) < min_depth:
        seq = 'N'
        return seq, af_low

    # If no indels, just report most seen and the frequency
    if (x.find('+') < 0) and (x.find('-') < 0):
        chars = ('*','A','C','T','G','N')
        counts = [x.count(y) for y in chars]
        seq = chars[counts.index(max(counts))]
        
        # If deletion is most common, insert no sequence
        if seq == '*':
            seq = ''
   
        # If allele frequency is low, replace with N
        if max(counts)/len(x) < min_af:
            seq = 'N'
            af_low = 1

    # If indels present, traverse through string
    else:

        # Initialize variables
        char_dict = {}
        deletion = False
        insertion = False
        del_count = 0
        ins_count = 0
        ins_finished = False
        s_prev = ''

        for s in x:
            # If in a deletion, skip this char
            if del_count > 0:
                del_count -= 1
                continue

            # If in an insertion, combine characters
            if ins_count > 0:
                ins_count -= 1
                ins_seq = ins_seq + s
                
                # When insertion finished, overwrite the s var
                if ins_count == 0:
                    s = s_prev + ins_seq
                    ins_finished = True
                else:
                    continue

            # If minus was detected in previous char, grap the quantifier
            if deletion:
                del_count = int(s)
                deletion = False
                continue

            # If plus was detected in previous char, grap the quantifier
            if insertion:
                ins_count = int(s)
                insertion = False
                ins_seq = ''
                continue 
            
            # If indel, enter new states
            if s == '-':
                deletion = True
            elif s == '+':
                insertion = True

            # If not in indel save in dict
            else:
                if s in char_dict:
                    char_dict[s] += 1
                else:
                    char_dict[s] = 1

                # If end of insertion, remove count of previous
                if ins_finished:
                    char_dict[s_prev] -= 1
                    ins_finished = False
                
                # Save previous char for insertions
                s_prev = s

        counts = list(char_dict.values())
        seq = list(char_dict.keys())[counts.index(max(counts))]
        
        # If deletion most common, insert no sequence
        if seq == '*':
            seq = ''

        # If allele frequency low, replace with N
        if max(counts)/len(x) < min_af:
            seq = 'N'
            af_low = 1

    return seq, af_low


def mpileup_read(file, min_af = const_min_af, min_depth = const_min_depth):
    '''
    Read a samtools mpileup file and output sequence dict

    min_af: Minimum allowed allele frequency, if above this an N is returned at that position

    Return:
    dict: key, value = sequence ID, [sequence, first position, last position, total depth, n bases with too low allele freq, seq_len]
    
    '''

    with open(file, 'r') as file_handle:

        # Initialize dict
        seq_dict = {}

        # Initialize variables for first line
        seq_id = None
        first = 0
        last = 0
        depth = 0
        af_count = 0
        seq_len = 0

        # Read data
        for line in file_handle:
                
            line_data = line.rstrip().split('\t')
            
            # Starting with new reference
            if line_data[0] != seq_id:
               
                # If this is first line in file, don't save
                if seq_id is None:
                    pass
                
                # If not first line in file, save results
                else:
                    seq_dict[seq_id] = [sequence, first, last, depth, af_count, seq_len]
                
                # Start variables for new reference
                seq_id = line_data[0]
                first = 0
                last = 0
                
                if int(line_data[3]) == 0:
                    seq_base = ''
                    af_ind = 0
                else:
                    seq_base, af_ind = parse_base_pairs(line_data[2], line_data[4], min_af, min_depth)
                
                sequence = seq_base
                af_count = af_ind

                seq_len = len(seq_base)
                depth = int(line_data[3])*len(seq_base)
         
                # Only save first if not a deletion
                if len(seq_base) > 0: 
                    first = int(line_data[1])
                else:
                    first = 0

            # Continuing with same reference as previous line
            else:
                if int(line_data[3]) == 0:
                    seq_base = ''
                    af_ind = 0
                else: 
                    seq_base, af_ind = parse_base_pairs(line_data[2], line_data[4], min_af, min_depth)
                
                sequence += seq_base
                af_count += af_ind

                seq_len += len(seq_base)
                depth += int(line_data[3])*len(seq_base)

                # If this is the first character
                if first == 0:
                    if len(seq_base) > 0:
                        first = int(line_data[1])

                # Save base if not deletion
                if len(seq_base) > 0:
                    last = int(line_data[1])

        # End of file, save results
        if len(sequence) > 0: 
            seq_dict[seq_id] = [sequence, first, last, depth, af_count, seq_len]

        return seq_dict



def read_refs(ref_fasta):
    '''
    Load the reference sequences to get the lengths
    '''
    len_dict = {}
    with open(ref_fasta, 'r') as fh:
        for fa in SeqIO.parse(fh, 'fasta'):
            len_dict[fa.id] = len(fa.seq)

    return(len_dict)

def read_roots(root_fasta):
    '''
    Load the root marker gene sequences
    '''
    seq_dict = {}
    with open(root_fasta, 'r') as fh:
        for fa in SeqIO.parse(fh, 'fasta'):
            seq_dict[fa.id] = str(fa.seq)

    return(seq_dict)

#######################################################################################################################
if __name__ == '__main__':
    main()



import os
import sys
import re
import logging

from Bio import SeqIO

'''
Filter bins by size
1st input VAMB clusters.tsv file
2nd input contigs fasta file
3rd input output directory for writing a fasta per cluster
4th input is bin size limit
'''

# Read VAMB file into dict
with open(snakemake.input[0], 'r') as fh:
    vamb_dict = [x.strip().split() for x in fh.readlines()]
vamb_dict = {x[1]:x[0] for x in vamb_dict}

# Read contigs and organize into dict per bin
bin_contigs = dict()
with open(snakemake.input[1], 'r') as fh:
    for fa in SeqIO.parse(fh, 'fasta'):
        bin_id = vamb_dict[fa.id]
        try:
            bin_contigs[bin_id].append(fa)
        except Exception:
            bin_contigs[bin_id] = [fa]

# Filter by length
bin_len = {x: sum([len(y.seq) for y in bin_contigs[x]]) for x in bin_contigs}
bin_filter = [x for x in bin_len if bin_len[x] >= int(snakemake.params.binsize)]

# Write per cluster
cluster_set = set([int(re.sub('.*_', '', x)) for x in bin_filter])

try:
    os.mkdir(snakemake.output[0])
except Exception:
    pass

for cluster in cluster_set:
    bins = [x for x in bin_filter if bool(re.search('_'+str(cluster)+'$', x))]
    clust_path = os.path.join(snakemake.output[0], str(cluster))

    try:
        os.mkdir(clust_path)
    except Exception:
        pass

    for bin_id in bins:
        SeqIO.write(bin_contigs[bin_id], os.path.join(snakemake.output[0], str(cluster), bin_id+'.fa'), "fasta")




import sys
import yaml
import re
import os
import subprocess
import logging
import glob

from maginator.messages import Message

class Workflow(object):

    def __init__(self, obj):
        for key, val in vars(obj).items():
            setattr(self, key, val)
        
    def add_info(self, x):
        
        # Substitute resource information
        x = re.sub('{cores}', '{resources.cores}', x)
        x = re.sub('{mem_gb}', '{resources.mem_gb}', x)
        x = re.sub('{runtime}', '{resources.runtime}', x)
   
        return x

    def run(self, snakefile):

        # Start message instance
        messages = Message(self)

        # Define core snakemake command
        cmd = ['snakemake',
               '--use-conda',
               '--latency-wait', '150',
               '-s', snakefile,
               '--config',
               'wd='+self.output,
               'reads='+self.reads,
               'contigs='+self.contigs,
               'vamb='+self.vamb_clusters,
               'params='+self.params]
        
        # If run on server
        if self.cluster == None:
            cmd += ['--cores', str(self.max_cores)]

        # If run on a cluster
        else:
            
            # Make logging dirs
            if self.cluster in ('qsub', 'slurm'):

                try:
                    os.mkdir(self.output + 'logs/cluster_err')
                except FileExistsError:
                    pass
                try:
                    os.mkdir(self.output + 'logs/cluster_out')
                except FileExistsError:
                    pass
            else:
                try:
                    os.mkdir(self.output + 'logs/drmaa')
                except FileExistsError:
                    pass
            
            # Add cluster info to snakemake command
            cmd += ['--jobs', str(self.max_jobs),
                    '--local-cores', str(self.max_cores)]
                
        if self.cluster == 'qsub':
            
            cluster_cmd = 'qsub' + ' ' + self.cluster_info
            cluster_cmd = cluster_cmd + ' -e ' + self.output + 'logs/cluster_err' + ' -o ' + self.output + 'logs/cluster_out'
            cluster_cmd = self.add_info(cluster_cmd)

            # Final snakemake command
            cmd += ['--cluster', cluster_cmd]

        if self.cluster == 'slurm':
            
            cluster_cmd = 'sbatch' + ' ' + self.cluster_info
            cluster_cmd = cluster_cmd + ' -e ' + self.output + 'logs/cluster_err/%j.err' + ' -o ' + self.output + 'logs/cluster_out/%j.out'
            cluster_cmd = self.add_info(cluster_cmd)
            
            # Final snakemake command
            cmd += ['--cluster', cluster_cmd]

        if self.cluster == 'drmaa':
            
            cluster_cmd = self.add_info(cluster_cmd)
            
            # Final snakemake command
            cmd += ['--cluster', cluster_cmd,
                    '--drmaa-log-dir', self.output+'logs/drmaa']
        
        # Only install conda envs if only_conda
        if self.only_conda:
            logging.info('Only creating conda environments.')
            cmd.append('--conda-create-envs-only')

        # If unlocking
        if self.unlock:
            logging.info('Unlocking working directory.')
            cmd.append('--unlock')

        logging.debug(' '.join(cmd))
        # Start snakemake process and read stdout and stderr (also save in logger)
        process = subprocess.Popen(cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE, text=True)
        if self.log_lvl == 'DEBUG':
            for line in iter(process.stdout.readline, ""):
                logging.debug(line.strip())
        process.wait()
        logging.debug('Snakemake returncode: '+str(process.returncode))

        # Check returncode
        if process.returncode != 0:
            for line in iter(process.stdout.readline, ""):
                logging.error(line.strip())
            logging.error('Snakemake returncode: '+str(process.returncode))
            sys.exit()

        # Save information on what has been run
        messages.add(snakefile) 

        # If unlocking exit program
        if self.unlock:
            sys.exit()

        # Print info
        if not self.only_conda:
            messages.check()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="maginator", 
    version="0.1.20",
    author="Jakob Russel & Trine Zachariasen",
    author_email="russel2620@gmail.com,trine_zachariasen@hotmail.com",
    description="MAGinator: Abundance, strain, and functional profiling of MAGs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Russel88/MAGinator",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Development Status :: 4 - Beta"],
    python_requires='>=3.5',
    install_requires=[
        "setuptools"],
    entry_points = {
        'console_scripts': ['maginator = maginator.main:cli']
    },
    include_package_data=True,
    zip_safe=False,
    package_data = {'maginator': ['workflow/*']}
)
